;;;; crunch - command-line driver


(module main ()

  (import (scheme base)
          (scheme read)
          (chicken base)
          (chicken pretty-print)
          (chicken port)
          (chicken condition)
          (chicken process-context)
          (chicken foreign))
  (import matchable srfi-1)
  (import (crunch compiler)
          (crunch compiler node)
          (crunch compiler util)
          (crunch compiler type))

  ;; directories

  (define crunch-cflags
    "-Wno-parentheses-equality -Wno-unused-value -fno-strict-aliasing -fwrapv")

  (define object-directory
    (or (get-environment-variable "CRUNCH_DIR")
        (get-environment-variable "CHICKEN_INSTALL_REPOSITORY")
        (and-let* ((dir (get-environment-variable "CHICKEN_INSTALL_PREFIX")))
          (string-append dir "/lib/chicken"
                         (number->string (foreign-value "C_BINARY_VERSION" int))))
        (foreign-value "C_INSTALL_EGG_HOME" c-string)))

  (define include-directory
    (or (and-let* ((dir (get-environment-variable "CHICKEN_INSTALL_PREFIX")))
          (string-append dir "/include"))
        (foreign-value "C_INSTALL_INCLUDE_HOME" c-string)))

  ;; scheme wrapper generation

  (define (emit-scheme-wrappers/module mname exps out)
    (pp `(module ,mname
           ,(map car exps)
           (import (scheme base) (chicken foreign))
           ,@(map (match-lambda
                    ((n rn t)
                     `(define ,rn ,(foreign-call-wrapper t n))))
                exps))
        out))

  (define (emit-scheme-wrappers/exported exps out)
    (pp `(begin
           ,@(map (match-lambda
                    ((n rn t)
                      `(define ,rn ,(foreign-call-wrapper t n))))
               exps))
        out))

  (define (emit-wrappers out)
    (lambda (mod exps)
      (if mod
          (emit-scheme-wrappers/module (##sys#module-name mod) exps out)
          (emit-scheme-wrappers/exported exps out))))

  (define short-options (string->list "wxvdJ"))

  (define (usage status)
    (with-output-to-port (current-error-port)
      (lambda ()
        (print "usage: crunch [FILENAME|OPTION] ...

  -help                      show this text
  -I DIRECTORY               add DIRECTORY to include path
  -w -whole-program          compile all files as a single unit
  -begin CODE                add code to compile
  -x -explain                explain optimizations
  -cflags                    print recommended C flags
  -libs                      print recommended linker flags
  -v -verbose                print progress
  -dump-expanded             dump expanded node tree
  -dump-resolved             dump typed node tree
  -dump-optimized            dump optimized node tree
  -dump-all                  dup node tree after every pass
  -d -debug                  enable debugger
  -list-types                list exports with their types
  -j -emit-import-library MODULENAME
                             emit import library for given module
  -J -emit-all-import-libraries
                             emit import libraries for all modules
  -emit-wrappers FILENAME    emit wrapper file to FILENAME
  -o -output-file FILENAME   write generated C code to FILENAME
  --                                                    treat remaining arguments as filenames")
        (exit status))))

  (define (string->expr str)
    (let rec ((in (open-input-string str)))
      (let ((x (read in)))
        (if (eof-object? x)
            '()
            (cons x (rec in))))))

  (define (main args)
    (let ((hascode #f)
          (flags #f))
      (let loop ((args args) (cargs '()))
        (match args
          (()
           (handle-exceptions ex
             (if (crunch-condition? ex)
                 (exit 1)
                 (abort ex))
             (if (or (null? cargs) (not hascode))
                 (if flags
                         (exit)
                         (usage 1))
                 (apply crunch cargs))))
          (("-cflags" . more)
            (print crunch-cflags " -I" include-directory)
            (set! flags #t)
            (loop more cargs))
          (("-libs" . more)
            (print (string-append object-directory "/crunch-utf.o")
                   " -lm")
            (set! flags #t)
            (loop more cargs))
          (("-begin" code . more)
            (set! hascode #t)
            (loop more
                  (cons (cons 'begin (string->expr code)) cargs)))
          (((or "-whole-program"
                "-explain"
                "-emit-all-import-libraries"
                "-verbose"
                "-dump-expanded"
                "-dump-resolved"
                "-dump-optimized"
                "-dump-all"
                "-debug"
                "-list-types") . more)
           (loop more (cons (string->symbol (substring (car args) 1)) cargs)))
          (("-w" . more)
           (loop more (cons 'whole-program cargs)))
          (("-x" . more)
           (loop more (cons 'explain cargs)))
          (("-d" . more)
           (loop more (cons 'debug cargs)))
          ((("-emit-wrappers" fname) . more)
           (set! wrapperfile fname)
           (loop more (cons (list 'export-consumer
                                  (emit-wrappers (open-output-file fname)))
                            cargs)))
          (("-v" . more)
           (loop more (cons 'verbose cargs)))
          (("-J" . more)
           (loop more (cons 'emit-all-import-libraries cargs)))
          (((or "-j" "-emit-import-library") name . more)
           (loop more (cons (list 'emit-import-library name) cargs)))
          (((or "-h" "-help" "--help") . _)
           (usage 0))
          (("-I" dir . more)
           (set! ##sys#include-pathnames (cons dir ##sys#include-pathnames))
           (loop more cargs))
          (((or "-o" "-output-file") fname . more)
           (loop more (cons (list 'output-file fname) cargs)))
          (("--" args ...)
           (loop '() (append args cargs)))
          ((str args ...)
           (set! hascode #t)
           (if (and (> (string-length str) 1)
                    (char=? #\- (string-ref str 0)))
               (let ((ss (string->list (substring str 1 (string-length str)))))
                 (if (null? (lset-difference char=? ss short-options))
                     (loop (append (map (cut string #\- <>) ss) args) cargs)
                     (usage 1)))
               (loop args (cons str cargs))))
          (_ (usage 1))))))

  (main (command-line-arguments))

)
