;;; crunch - type inference


(module (crunch compiler inference) (analyze)

  (import (scheme base)
          (scheme write)
          (scheme cxr)
          (scheme file)
          srfi-1
          matchable
          (chicken plist)
          (chicken base)
          (chicken pretty-print)
          (chicken memory representation)
          (chicken format)
          (chicken number-vector))

  (import (crunch compiler util)
          (crunch compiler node)
          (crunch compiler type))

  (define analysis-errors (make-parameter '()))

  (define (type-error n msg . args)
    (let* ((errs (analysis-errors))
           (loc (node-location n))
           (e (list loc msg args)))
      (unless (member e errs)
        (analysis-errors (cons e errs)))
      #f))

  (define (report-errors)
    (for-each
      (match-lambda
        ((loc msg args)
          (elucidate
            (if loc
                (format "Error: ~a: ~?\n" loc msg args)
                (format "Error: ~?\n" msg args)))))
      (reverse (analysis-errors))))

  (define (unknown! n) (type-node! n (make-typevar '? '(?))))

  (define (type-node! n t)
    (node-type-set! n t)
    t)

  (define (expr->node x loc where)
    (let walk ((x x) (e '()))
      (match x
        ((or (? number?) (? vector?) (? string?) (? char?))
          (node* loc 'const (list x)))
        (('quote x)
          (node* loc 'const (list x)))
        (('if t c a)
          (node* loc 'if '() (walk t e) (walk c e) (walk a e)))
        (('begin xs ...)
          (apply node* loc 'begin '() (map (cut walk <> e) xs)))
        ((? symbol?)
          (cond ((assq x e) =>
                  (lambda (a)
                    (node* loc 'local-ref (list (cdr a) x))))
                (else (node* loc 'global-ref (list x #f x)))))
        (('let bs b1 b2 body ...)
          (walk `(let ,bs (begin ,b1 ,b2 ,@body)) e))
        (('let bs body)
          (let ((as (map (lambda _ (gensym 'v)) bs)))
            (apply node* loc 'let
              (list (map car bs) as)
              (append (map (lambda (b) (walk (cadr b) e)) bs)
                      (list (walk body (append (map (lambda (b a) (cons (car b) a)) bs as)
                      			       e)))))))
        (('set! v x)
          (node* loc 'local-set! (list (cdr (assq v e)) v) (walk x e)))
        (('lambda llist x)	; no #!optional allowed
          (let ((as (map (lambda _ (gensym 'v)) llist))
                (id (gensym 'f)))
            (node* loc 'lambda (list id (length llist) llist as where)
              (walk x (append (map (lambda (v a) (cons v a)) llist as) e)))))
        (('unspecified) (node* loc 'unspecified '()))
        ((fn args ...)
          (apply node* loc 'call '()
            (map (cut walk <> e) (cons fn args))))
        (_ (error 'expr->node "bad inline expression" x)))))

  ;; initial type decoration walk
  (define (decorate n)
    (let ((lambdas '()))
      (define (lambda-type ln)
        (set! lambdas (cons ln lambdas))
        (put! (first (node-data ln)) 'crunch:lambda-literal #t)
        (node-type ln))
      (define (value-type x n)
        (match (literal-type x lambda-type)
          (('error fmt args ...)
           (apply type-error n fmt args)
           '?)
          (x x)))
      (define (walk n e tl? #!optional main?) ; returns result type
        (case (node-class n)
          ((const)
           (let ((x (first (node-data n))))
             (cond ((not (procedure? x))
                    (type-node! n (make-typevar (value-type x n) (value-set x))))
                   ((and (extended-procedure? x)
                         (match (procedure-data x)
                           (('crunch:lambda ln) ln)
                           (_ #f))) =>
                    (lambda (ln)
                      (let* ((id (first (node-data ln)))
                             (seen (get id 'crunch:instantiated)))
                        (cond (seen
                                (set-car! (node-data n) id)
                                (type-node! n (get id 'crunch:lambda)))
                              (else
                                (let ((t (walk ln e tl? main?)))
                                  (put! id 'crunch:instantiated #t)
                                  (put! id 'crunch:lambda-literal #t)
                                  (node-class-set! n 'begin)
                                  (node-data-set! n '())
                                  (node-subnodes-set!
                                    n
                                    (list ln
                                          (make-node
                                            (node-location n)
                                            'const (list id) '() t)))
                                  (type-node! n t)))))))
                   (else
                     (type-error n "can not compile host procedure: ~s" x)))))
          ((global-ref)
           (match (node-data n)
             ((var _ _)
              (cond ((get var 'crunch:type) =>
                     (lambda (t)
                       (let ((t (clone-typevar t))
                             (callid (get var 'crunch:function-id)))
                         (if callid
                             (typevar-value-set-set! t (list callid))
                             (match (pure-value-set t)
                               ((id)
                                ;; imported
                                (put! id 'crunch:extern t))
                               (_ #f)))
                         (type-node! n t))))
                    ((get var 'crunch:inline) =>
                     (lambda (inl)
                       (let ((proc (if (procedure? inl)
                                       (inl n)
                                       (expr->node inl (node-location n) var))))
                         (replace-node! n proc)
                         (walk n e #f))))
                    (else
                      (let ((gt (make-typevar '? '(?))))
                        (put! var 'crunch:type* gt)
                        (type-node! n gt)))))))
          ((local-ref)
           (type-node! n (cdr (assq (first (node-data n)) e))))
          ((unspecified) (unknown! n))
          ((if)
           (match (node-subnodes n)
             ((tst cns alt)
              (let ((_ (walk tst e #f))
                    (tc (walk cns e tl?))
                    (ta (walk alt e tl?)))
                ;; void explicitly unspecified branches
                (when (eq? 'unspecified (node-class cns))
                  (type-node! cns (void-type)))
                (when (eq? 'unspecified (node-class alt))
                  (type-node! alt (void-type)))
                (unknown! n)))))        ; done later
          ((begin)
           (let loop ((ns (node-subnodes n)))
             (match ns
               ((tn) (type-node! n (walk tn e tl? main?)))
               ((sn . more)
                (walk sn e #f)
                (loop more)))))
          ((global-set!)
           (match (node-data n)
             ((var exp? lname)
              (let ((vt (walk (first (node-subnodes n)) e #f
                              (and exp? (eq? lname 'main))))
                    (gt (get var 'crunch:type))
                    (dt (get var 'crunch:declared-type)))
                (unless gt
                  (put! var 'crunch:type* vt))
                (when dt
                  (let ((set (pure-value-set vt)))
                    (unless (unify-types! vt dt (vector #f))
                      (type-error
                        n "declared type does not match for variable `~a': ~s"
                        var (typename dt)))
                    ;; preserve value set or it will be destroyed by
                    ;; the unification, we need this for exporting.
                    (typevar-value-set-set! vt set)
                    (put! var 'crunch:type* vt)))
                (match (pure-value-set vt)
                  ((x)
                   (unless (eq? '? x)
                     (put! var 'crunch:initial-value x)
                     (when (and (symbol? x) exp?)
                       (put! x 'crunch:escapes #t))))
                  (_ #f))
                (type-node! n (void-type))))))
          ((local-set!)
           (walk (first (node-subnodes n)) e #f)
           (let ((a (assq (first (node-data n)) e)))
             ;; extend node data to hold variable type
             (node-data-set! n (append (node-data n) (list (cdr a))))
             (type-node! n (void-type))))
          ((let)
           (match (node-data n)
             ((vars aliases)
              (let* ((ns (node-subnodes n))
                     (e2 (append (map (lambda (a x) (cons a (walk x e #f)))
                                   aliases (butlast ns))
                                 e)))
                (type-node! n (walk (last ns) e2 tl? main?))))))
          ((lambda)
           (match (node-data n)
             ((id 'values (var) (alias) where)
              (let* ((e2 (alist-cons alias
                                     (make-typevar `(values ,(make-typevar '? '(?)))
                                                   '())
                                     e))
                     (bt (walk (first (node-subnodes n)) e2 #t))
                     (pt (type-node! n (make-typevar (procedure-type/argc
                                                       1
                                                       (list (cdar e2))
                                                       bt)
                                                     (list id)))))
                (set-car! (cdr (node-data n)) 1)
                (typevar-value-set-set! pt (list id))
                (put! id 'crunch:lambda pt)
                (put! id 'crunch:instantiated #t)
                pt))
             ((id argc vars aliases where)
              (when main?
                (set! argc 3)
                ;; no need to mark vars as used, as main escapes
                (match aliases
                  ((_ _)
                   (let ((ae (gensym 'env)))
                     (set! aliases (append aliases (list ae)))
                     (node-data-set!
                       n
                       (list id 3 (append vars (list ae)) aliases where))))
                  ((_)
                   (let ((ve (gensym 'argv))
                         (ae (gensym 'env)))
                     (set! aliases (append aliases (list ve ae)))
                     (node-data-set!
                       n
                       (list id 3 (append vars (list ve ae)) aliases where))))
                  (()
                   (let ((ce (gensym 'argc))
                         (ve (gensym 'argv))
                         (ae (gensym 'env)))
                     (set! aliases (append aliases (list ce ve ae)))
                     (node-data-set!
                       n
                       (list id 3 (append vars (list ce ve ae)) aliases where))))
                  (_ #f))) ;XXX error?
              (let* ((le (if main?
                             (map (lambda (a t) (cons a (typespec t)))
                               aliases
                               '(integer
                                  (pointer (pointer char))
                                  (pointer (pointer char))))
                             (map (lambda (a) (cons a (make-typevar '? '(?))))
                               aliases)))
                     (e2 (append le e))
                     (bt (walk (first (node-subnodes n)) e2 #t))
                     (pt (type-node! n (make-typevar (procedure-type/argc
                                                       argc
                                                       (map cdr le)
                                                       bt)
                                                     (list id)))))
                (typevar-value-set-set! pt (list id))
                (put! id 'crunch:lambda pt)
                (put! id 'crunch:instantiated #t)
                pt))))
          ((call)
           (node-data-set! n (list tl?))
           (match (node-subnodes n)
             ((fn . args)
              ;; do this like global-ref, but explicitly to avoid walking the body twice
              ;; (we need this here to perform the walk with the proper environment)
              (and-let* ((_ (eq? 'global-ref (node-class fn)))
                         (var (first (node-data fn)))
                         (inl (get var 'crunch:inline)))
                (let ((proc (if (procedure? inl)
                                (inl n)
                                (expr->node inl (node-location n) var))))
                  (replace-node! fn proc)))
              (cond ((eq? (node-class fn) 'lambda)
                     (match (node-data fn)
                       ((_ _ vs as _)
                        (node-class-set! n 'let)
                        (node-data-set! n (list vs as))
                        ;; assumes no #!optional
                        (node-subnodes-set!
                          n
                          (let rec ((vs vs) (args args))
                            (cond ((null? vs)
                                   (unless (null? args)
                                     (type-error n "argument count to inlined lambda exceeds expected number"))
                                   (node-subnodes fn))
                                  ((null? args)
                                   (type-error n "missing arguments in call to inlined lambda"))
                                  (else (cons (car args) (rec (cdr vs) (cdr args)))))))
                        (walk n e tl?))))
                    (else
                      (map (cut walk <> e #f) (node-subnodes n))
                      (unknown! n))))))        ; done later
          ((c-external)
           (match (node-data n)
             ((inc name args ret)
              (let ((ct (make-typevar `(procedure ,(map (lambda (t)
                                                          (if (eq? '#!optional t)
                                                              t
                                                              (typespec t)))
                                                     args)
                                                  ,(typespec ret))
                                      (list name))))
                (type-node! n ct)
                (put! name 'crunch:external-name name)
                (put! name 'crunch:escapes #t) ; to avoid integration
                (put! name 'crunch:lambda ct)
                ct))))
          ((c-lambda)
           (match (node-data n)
             ((name ret args body)
              (let ((ct (make-typevar `(procedure ,(map (lambda (arg) (typespec (car arg)))
                                                     args)
                                                  ,(typespec ret))
                                      (list name))))
                (type-node! n ct)
                (put! name 'crunch:escapes #t) ; to avoid integration
                (put! name 'crunch:lambda ct)
                ct))))
          ((c-declare)
           (type-node! n (void-type))
           (node-type n))
          (else => (cut error 'walk "invalid node class" <>))))
      (walk n '() #f)
      (let loop ((lns '()))
        (let ((lns2 (filter-map
                      (lambda (ln)
                        (and (not (get (first (node-data ln)) 'crunch:instantiated))
                             ln))
                      lambdas)))
          (cond ((null? lns2)
                 (let ((t (void-type)))
                   (make-node
                     (node-location n) 'begin '()
                     (append
                       lns
                       (list n
                             (make-node (node-location n) 'unspecified '()
                                        '() t)))
                     t)))
                (else
                  (set! lambdas '())
                  (for-each
                    (lambda (n)
                      (walk n '() #f)
                      (put! (first (node-data n)) 'crunch:lambda-literal #t))
                    lns2)
                  (loop (lset-union eq? lns2 lns))))))))

  (define (solve n cell mf)
    (case (node-class n)
      ((global-ref)
        (let* ((var (first (node-data n)))
               (vt (get var 'crunch:type*)))
          (when vt (unify-types! (node-type n) vt cell))))
      ((if)
        (match (node-subnodes n)
          ((tst cns alt)
            (solve tst cell mf)
            (solve cns cell mf)
            (solve alt cell mf)
            (unify-types! (node-type n)
                          (merge-types (node-type cns) (node-type alt))
                          cell))))
      ((lambda)
        (match (node-data n)
          ((id argc _ aliases where)
           (let ((body (first (node-subnodes n))))
             (solve body cell mf)
             (let ((pt (node-type n)))
               (unify-types! (node-type n)
                             (make-typevar
                               `(procedure ,(procedure-argument-types pt)
                                           ,(merge-types (procedure-result-type pt)
                                                         (node-type body)))
                               (list id))
                             cell))))))
      ((global-set!)
        (let* ((v (first (node-subnodes n)))
               (var (first (node-data n)))
               (lt (get var 'crunch:type*)))
          (solve v cell mf)
          (when lt
            (unify-types! lt (node-type v) cell))))
      ((local-set!)
        (let ((v (first (node-subnodes n))))
          (solve v cell mf)
          (unify-types! (third (node-data n)) (node-type v) cell)))
      ((idcall)
       (for-each (cut solve <> cell mf) (node-subnodes n))
       (let* ((id (first (node-data n)))
              (pt (get id 'crunch:lambda))
              (pt2 (make-typevar
                     (or (procedure-type/optionals
                           pt
                           (map node-type (node-subnodes n))
                           (node-type n))
                         (error "can not match procedure type" (typename pt)
                                (node-shape n)))
                     (list id))))
         (unify-types! pt pt2 cell)
         (let ((rt (procedure-result-type pt)))
           (unify-types! (node-type n) rt cell))))
      ((call)
        (match (node-subnodes n)
          ((fn . args)
           (solve fn cell mf)
           (for-each (cut solve <> cell mf) args)
           (let* ((fnt (node-type fn))
                  (vset (pure-value-set fnt))
                  (argts (procedure-argument-types fnt))
                  (opts (and argts (memq '#!optional argts)))
                  (tl (first (node-data n)))
                  (rts (procedure-result-type fnt))
                  (uni #t)
                  (ids '()))
             (for-each
               (lambda (v)
                 (when (symbol? v)
                   (cond ((and (= mf 1)
                   			   (get v 'crunch:variant)) =>
                          (match-lambda
                            ((varn lamn)
                             ;; clone variant
                             (receive (new id)
                                 (find-variant varn lamn (map node-type args))
                               (when id
                               	 ;; if fnt is #f, it refers to a variant,
                               	 ;; just override types (crock)
                                 (unless argts
                     			   (set! fnt (clone-typevar (node-type lamn)))
                                   (set! argts (procedure-argument-types fnt))
                                   (set! rts (procedure-result-type fnt)))
                                 (set! uni #f)
                                 (when ids
                                   (set! ids (cons id ids))))))))
                         ((and (= mf 1)
                               (get v 'crunch:lambda)) =>
                          ;; associate lambda type with call
                          (lambda (pt)
                            (let ((pt2 (make-typevar (procedure-type/optionals
                                                       pt
                                                       (map node-type args)
                                                       (node-type n))
                                                     (list v))))
                              (unify-types! pt pt2 cell)
                              (when ids (set! ids (cons v ids)))
                              (set! uni #f)
                              (and-let* ((rt (procedure-result-type pt)))
                                (unify-types! (node-type n) rt cell)))))
                         ((get v 'crunch:overload) =>
                          ;; apply overloading
                          (lambda (ol)
                            (let* ((ats (map node-type args))
                                   (ots (ol (map typename ats))))
                              (set! ids #f)
                              (set! uni #f)
                              (match ots
                                ((atss rts)
                                 (let* ((rt (typespec rts))
                                        (pt1 (node-type fn))
                                        (pt (make-typevar
                                              (procedure-type/optionals
                                                pt1
                                                (map typespec atss)
                                                rt)
                                              (list v))))
                                   (unify-types! pt1 pt cell)
                                   (unify-types! (node-type n) rt cell)))
                                (_ #f))))))))
               vset)
             ;; when not overloaded or variant or lambda, just unify argument
             ;; types and result
             (when (and (> mf 1) uni argts)
               (unify-argtypes! argts (map node-type args) cell)
               (unify-types! rts (node-type n) cell))
             (match ids
               ((id)
                 (touch! cell)
                 (replace-with-idcall! n fn args id opts #f tl))
               ((id1 ids ...)
                 (touch! cell)
                 (replace-with-dispatch! n fn args (cons id1 ids) opts #f tl))
               (_ #f))
             (when rts
               (unify-types! (node-type n) rts cell))))))
    ((unspecified)
      ;; now set the type, as remaining nodes of this class are (probably)
      ;; not placeholders anymore
      (when (= mf 3)
        (unify-types! (node-type n) (void-type) cell)))
    ((variant)
      (when (= mf 2)
        (let* ((tv (node-type n))
               (id (first (typevar-value-set tv))))
          (match (get id 'crunch:variant)
            (#f #f)
            ((varn _)
             ;; select first variant lambda
             (let ((ns (node-subnodes varn)))
               (when (and (not tv) (pair? ns))
                 (type-copy! tv (node-type (first ns)))
                 ;; drop id's from variant's original
                 (typevar-value-set-set!
                   tv
                   (list (first (node-data (first ns)))))
                 (touch! cell)))))))
      (for-each (cut solve <> cell mf) (node-subnodes n)))
    (else
      (for-each (cut solve <> cell mf) (node-subnodes n)))))

  (define (check-type n)
    (let* ((tv (node-type n))
           (t (typevar-type tv))
           (vset (typevar-value-set tv)))
      (cond ((eq? '? t)
              (type-error n "~a `~s' has unknown type"
                (case (node-class n)
                  ((local-ref) "local variable")
                  ((global-ref) "global variable")
                  ((call idcall) "procedure call")
                  (else "expression"))
                (node-shape n)))
            ((not (procedure-type? tv)))
            ((and (null? vset) ; if value(s) exist, be more lax
                  (eq? '? (typevar-type (procedure-result-type tv))))
             (type-error n
                         "expression `~s' has procedure type with unknown result: ~s"
                         (node-shape n) (typename tv)))
            ((and (null? vset)
            	  (any (lambda (t)
                         (and (not (eq? t '#!optional))
                              (eq? (typevar-type t) '?)))
                   (procedure-argument-types tv)))
             (type-error n
                         "expression `~s' has procedure type with unknown argument: ~s"
                         (node-shape n) (typename tv)))
            (else #t))))

  (define (check n dest)
    (case (node-class n)
      ((if)
       (match (node-subnodes n)
         ((t c a)
          (let ((t (check t #f)))
            (and t
                 (check c dest)
                 (check a dest)
                 (check-type n))))))
      ((lambda)
       (cond ((procedure-type? (node-type n))
              (check (first (node-subnodes n)) #f))
             (else
               (type-error n
                           "lambda expression has non-procedure type `~s': ~s"
                           (typename (node-type n))
                           (node-shape n)))))
      ((local-set!)
        (and (check (first (node-subnodes n)) (first (node-data n)))
             (check-type n)))
      ((global-set!)
       (let* ((var (first (node-data n)))
              (val (first (node-subnodes n)))
              (gt (get var 'crunch:type))
              (lt (get var 'crunch:type*)))
         (and (check val var)
              (if (and gt (not (type>=? gt lt)))
                  (type-error
                    n
                    "assigned type for global variable `~a' in `~s' does not match declared type.\n value type: ~s\n global type: ~s"
                    var (node-shape n) (typename lt) (typename gt))
                  (let ((vs (pure-value-set lt)))
                    (put! var 'crunch:type lt)
                    (when (memq 'list-types (or (verbose) '()))
                      (cond ((and (pair? vs)
                                  (null? (cdr vs))
                                  (symbol? (car vs))
                                  (get (car vs) 'crunch:variant)) =>
                             (match-lambda
                               ((varn _)
                                (let ((numv (length (node-subnodes varn))))
                                  (printf "~a: (~a variant~a)\n" var
                                          numv (if (> numv 1) "s" "")))
                                (for-each
                                  (lambda (n)
                                    (printf "\t~s\n" (typename (node-type n))))
                                  (node-subnodes varn)))))
                            (else (printf "~a:\t~s\n" var (typename lt)))))
                    (match (and (second (node-data n)) ; exported?
                                (pure-value-set lt))
                      (((? symbol? id))
                       (put! id 'crunch:escapes #t)
                       (put! id 'crunch:external-name (third (node-data n))))
                      (_ #f))
                    (match (typevar-type lt)
                      (('procedure _ _) #t)
                      (_ (if (second (node-data n))
                             (type-error
                               n
                               "exported global variable `~a' holds non-procedure type: ~s"
                               var (typename lt))
                             #t)))
                    #t)))))
      ((call)
       (match (node-subnodes n)
         ((fn . args)
          (and (check fn #f)
               (every (cut check <> #f) args)
               (let* ((pt1 (node-type fn))
                      (pt2 (procedure-type/optionals
                             pt1
                             (map node-type args)
                             (node-type n)))
                      (pt (make-typevar pt2 '())))
                 (if (and pt2 (type>=? pt1 pt))
                     (wrap-call (canonicalize-call n) dest)
                     (type-error
                       n
                       "argument types in procedure call `~s' do not match.\n call: ~s\n expected: ~s"
                       (node-shape n)
                       (if pt2
                       	   (typename pt)
                       	   `(procedure ,(map (o typename node-type) args)
                       	      ,(typename (node-type n))))
                       (typename pt1))))))))
      ((idcall)
        (and (every (cut check <> #f) (node-subnodes n))
             (check-type n)
             (wrap-call n dest)))
      ((variant)
       (every (cut check <> dest) (node-subnodes n)))
      #;((local-ref)
       (print `(REF: ,(node-shape n) ,(typename (node-type n))));XXX
       (check-type n))
      (else
        (and (every (cut check <> #f) (node-subnodes n))
             (check-type n)))))

  (define (wrap-call n dest)
    (if (or dest (not (memq (node-type n) '(void noreturn))))
        n
        (replace-node!
          n
          (let ((tmp (gensym 'rt)))
            (make-node
              (node-location n)
              'let (list (list tmp) (list tmp))
              (list (clone-node n)
              	    (make-node
              	      (node-location n)
              	      'local-ref (list tmp tmp)
              	      '()
              	      (node-type n)))
              (node-type n))))))

  (define (canonicalize-call n)
    (match (node-subnodes n)
      ((fn . args)
       (let* ((fnt (node-type fn))
              (vset (pure-value-set fnt))
              (tl (first (node-data n)))
              (loc (node-location n))
              (argts (procedure-argument-types fnt))
              (rts (procedure-result-type fnt))
              (opts (memq '#!optional argts)))
         (cond ((null? vset) n)
               ((null? (cdr vset))
                 (replace-with-idcall! n fn args (car vset) opts #t tl))
               (else
                 (assert (every symbol? vset))        ; should always be true
                 (replace-with-dispatch! n fn args vset opts #t tl)))))))

  (define (replace-with-idcall! n fn args id opts? spec? tl?)
    (let ((tmp (gensym 't))
          (loc (node-location n))
          (nt (node-type n)))
      (replace-node!
        n
        (make-node
          loc 'let (list (list tmp) (list tmp))
          (list
            fn
            (bind-nontrivial
              args
              (lambda () (gensym 't))
              (lambda (args)
                (make-node
                  loc
                  'idcall
                  (list (if spec? (specialize id args) id)
                        tl? opts?)
                  args
                  nt))))
          nt))))

  (define (replace-with-dispatch! n fn args ids opts? spec? tl?)
    (let ((nt (node-type n))
          (loc (node-location n)))
      (replace-node!
        n
        (bind-nontrivial
          args
          (lambda () (gensym 't))
          (lambda (args)
            (let ((ids (if spec?
                           (map (cut specialize <> args) ids)
                           ids)))
              (make-node
                loc 'dispatch (list ids)
                (cons fn
                      (map (lambda (id)
                             (make-node loc
                                        'idcall
                                        (list id tl? opts?)
                                        args
                                        nt))
                        ids))
                nt)))))))

  (define (specialize id args)
    (cond ((get id 'crunch:specialize) =>
    		(lambda (sp)
    		  (cond ((sp (map (lambda (n) (typename (node-type n))) args)) =>
    		  	   (lambda (sid)
    		  	     ;; mark as primitive to avoid integration
    		  	     (when (symbol? sid) (put! sid 'crunch:primitive #t))
    		  	     sid))
    		  	(else id))))
          (else id)))

  (define (mark-monomorphization-candidates n)
    (let walk ((n n))
      (case (node-class n)
        ((lambda)
         (match (typevar-type (node-type n))
           (('procedure args _)
            (let ((id (first (node-data n))))
              (if (and (not (get id 'crunch:escapes))
                       (any (lambda (t)
                       	      (and (typevar? t)
                       	      	   (eq? '? (typevar-type t))))
                       	    args))
                  (let ((new (make-node (node-location n)
                                        'lambda
                                        (node-data n)
                                        (node-subnodes n)
                                        (node-type n))))
                    (dribble "monomorphization candidate ~a" id)
                    (node-class-set! n 'variant)
                    (node-data-set! n (list id))
                    (node-type-set! n (make-typevar #f (list id)))
                    (node-subnodes-set! n '())
                    (put! id 'crunch:variant (list n new)))
                  (walk (first (node-subnodes n))))
              (for-each walk (node-subnodes n))))
           (_ #f)))
        (else (for-each walk (node-subnodes n))))))

  (define (find-variant varn lamn argts)
    (let loop ((vs (node-subnodes varn)))
      (if (null? vs)
          (if (every known-type? argts)
              (let* ((nv (clone-lambda-node lamn argts))
                     (pt1 (node-type nv))
                     (id (first (node-data lamn)))
                     (nid (first (node-data nv)))
                     (pt2 (procedure-type/optionals
                            pt1
                            argts
                            (procedure-result-type pt1)))
                     (pt (make-typevar pt2 (list nid))))
                (explain lamn "cloning variant ~a of ~a for types ~s"
                         nid id (map typename argts))
                (node-subnodes-set! varn (cons nv (node-subnodes varn)))
                (put! nid 'crunch:variant-of nid)
                (put! nid 'crunch:lambda pt)
                (put! nid 'crunch:lambda-node nv)
                (values #t nid))
              (values #f #f))
          (match (typevar-type (node-type (car vs)))
            (('procedure vargs _)
             (if (every (lambda (vt at)
                          (or (eq? vt at)
                              (and (typevar? vt) (typevar? at)
                                   (type>=? vt at))))
                        vargs argts)
                 (values #f (first (node-data (car vs))))
                 (loop (cdr vs))))))))

  (define (explain n fstr . args)
    (when (memq 'explain (or (verbose) '()))
      (elucidate (format "~a: ~?\n" (or (node-location n) "<toplevel>")
        		   fstr args))))

  (define (clone-lambda-node n argts)
    (let ((lmap '()))  ; maps cloned inner lambda-ids
      (let clone ((n n) (e '()))
        (case (node-class n)
          ((unspecified const global-ref if global-set! call idcall)
           (make-node (node-location n)
                      (node-class n)
                      (list-copy (node-data n))
                      (map (cut clone <> e) (node-subnodes n))
                      (clone-typevar (node-type n) lmap)))
          ((begin)
           (let ((ns (map (cut clone <> e) (node-subnodes n))))
             (make-node (node-location n)
                        'begin
                        '()
                        ns
                        (node-type (last ns)))))
          ((local-ref)
           (match (node-data n)
             ((alias var)
              (let ((a (assq alias e)))
                (if a
                    (make-node (node-location n)
                               'local-ref
                               (list (second a) var)
                               '()
                               (third a))
                    (make-node (node-location n)
                               'local-ref
                               (list alias var)
                               '()
                               (node-type n)))))))
          ((local-set!)
           (match (node-data n)
             ((alias var tv)
              (let ((a (assq alias e)))
                (make-node (node-location n)
                           'local-set!
                           (if a
                               (list (second a) var (third a))
                               (list alias var tv))
                           (list (clone (first (node-subnodes n)) e))
                           (clone-typevar (node-type n) lmap))))))
          ((let)
           (let* ((ns (node-subnodes n))
                  (d (node-data n))
                  (vs (map (cut clone <> e) (butlast ns)))
                  (nas (map gensym (second d)))
                  (e2 (map (lambda (a na vn) (list a na (node-type vn)))
                        (second d)
                        nas
                        vs))
                  (b (clone (last ns) (append e2 e))))
             ;; adjust value sets for cloned inner lambda id's
             (for-each
               (lambda (a)
                 (typevar-value-set-set!
                   (third a)
                   (adjust-cloned-ids (typevar-value-set (third a)) lmap)))
               e2)
             (make-node (node-location n)
                        'let
                        (list (first d) nas)
                        (append vs (list b))
                        (node-type b))))
          ((lambda)
           (match (node-data n)
             ((id argc vs as h)
              (let* ((fid (gensym id))
                     (nas (map gensym as))
                     (e2 (if (null? e)  ; the actual cloned lambda?
                             (map list as nas argts)
                             (map (lambda (a na)
                                    (list a na (make-typevar '? '(?))))
                               as nas)))
                     (b (clone (first (node-subnodes n)) (append e2 e)))
                     (pt (make-typevar (procedure-type/argc
                                         argc
                                         (map third e2)
                                         (node-type b))
                                       (list fid))))
                (put! fid 'crunch:lambda pt)
                (set! lmap (alist-cons id fid lmap))
                (make-node (node-location n)
                           'lambda
                           (list fid argc vs nas h)
                           (list b)
                           pt)))))
          ((variant)
           ;;XXX is this right? we have to update crunch:variant info
           (make-node (node-location n)
                      'variant
                      (node-data n) ; XXX and this?
                      (map (cut clone <> e) (node-subnodes n))
                      #f))
          ;; dispatch not available yet
          (else =>
            (lambda (c) (error "unsupported node class while cloning" c)))))))

  (define (show n)
    (pp (node-shape n #f
          (lambda (n x) `(the ,(typename (node-type n)) ,x)))))

  (define (analyze n)
    (dribble "decorating tree")
    (set! n (decorate n))
    (let loop ((i 1) (mf 0))
      (dribble "type propagation pass #~a" i)
      (let ((cell (vector #f)))
        (solve n cell mf)
        (cond ((vector-ref cell 0) (loop (add1 i) mf))
              ((= mf 0)
               (dribble "enabling monomorphization")
               (mark-monomorphization-candidates n)
               (loop (add1 i) 1))
              ((= mf 1)
               (dribble "enabling variant resolution")
               (loop (add1 i) 2))
              ((= mf 2)
               (dribble "fixing void nodes")
               (loop (add1 i) 3)))))
    (dribble "checking types")
    (flush-output-port)
    ;(show n);XXX
    (cond ((check n #f) n)
          ((null? (analysis-errors)) n)
          (else
            (report-errors)
            ((node-dump-hook) n "ERROR: type checking failed")
            #f)))

)
