;;;; crunch - driver

(module (crunch compiler) (crunch crunch-version declare-entity)

  (import (scheme base)
          (scheme read)
          (scheme file)
          (chicken base)
          (chicken syntax)
          (chicken plist)
          (chicken pretty-print)
          (chicken condition)
          (chicken pathname)
          (chicken format)
          (chicken foreign)
          (chicken process)
          (chicken string)
          (chicken port)
          srfi-1
          matchable)

  (import (crunch compiler util)
          (crunch compiler expand)
          (crunch compiler inference)
          (crunch compiler node)
          (crunch compiler type)
          (crunch compiler primitives)
          (crunch compiler optimize)
          (crunch compiler backend)
          (crunch compiler debug))

  (define crunch-version (make-parameter 0.1))
  (define import-libraries (make-parameter '()))
  (define all-import-libraries (make-parameter #t))
  (define output-filename (make-parameter #f))
  (define consumer (make-parameter #f))
  (define export-consumer (make-parameter #f))
  (define c-compiler (foreign-value "C_TARGET_CC" c-string))
  (define compile-to-binary (make-parameter #f))

  (define (parse-args args k)
    (let loop ((args args) (fnames '()))
      (match args
        (() (k fnames))
        (('export-all . more)
         (register-exports 'all)
         (loop more fnames))
        (('whole-program . more)
         (whole-program #t)
         (loop more fnames))
        (('explain . more)
         (verbose (cons 'explain (or (verbose) '())))
         (loop more fnames))
        (('emit-all-import-libraries . more)
         (all-import-libraries #t)
         (loop more fnames))
        ((('emit-import-library name) . more)
         (import-libraries (cons name (import-libraries)))
         (loop more fnames))
        ((('output-file fname) . more)
         (output-filename fname)
         (loop more fnames))
        (('debug . more)
         (connect-to-debugger
           (lambda ()
             (parameterize ((information-handler
                              (lambda (m _) (emit-debug-message m)))
                            (node-dump-hook debug-node-tree)
                            (consumer (debug-consumer))
				            (verbose (or (verbose) '())))
               (loop more fnames)))))
        (('verbose . more)
         (verbose (or (verbose) '()))
         (loop more fnames))
        (('dump-all . more)
         (verbose (cons* 'expanded 'resolved 'optimized (or (verbose) '())))
         (loop more fnames))
        (('dump-expanded . more)
         (verbose (cons 'expanded (or (verbose) '())))
         (loop more fnames))
        (('list-types . more)
         (verbose (cons 'list-types (or (verbose) '())))
         (loop more fnames))
        ((('consumer proc) . more)
         (consumer proc)
         (loop more fnames))
        ((('export-consumer proc) . more)
         (export-consumer proc)
         (loop more fnames))
        (('dump-resolved . more)
         (verbose (cons 'resolved (or (verbose) '())))
         (loop more fnames))
        (('dump-optimized . more)
         (verbose (cons 'optimized (or (verbose) '())))
         (loop more fnames))
        ((('begin forms ...) more ...)
         (loop more (append fnames `((begin ,@forms)))))
        ((('compile-to-binary outfile options ...) more ...)
         (compile-to-binary (cons outfile options))
         (loop more fnames))
        ((('error-handler proc) more ...)
         (error-handler proc)
         (loop more fnames))
        ((('information-handler proc) more ...)
         (information-handler proc)
         (loop more fnames))
        (((? string? fname) . more)
         (loop more (append fnames (list fname))))
        (_ (error 'crunch "invalid argument" (car args))))))

  (define (dump n #!optional pass (types #t) (out (current-output-port)))
    (define (typed n x)
      (let ((t (typename (node-type n))))
        (if (eq? t '*) x `(the ,t ,x))))
    (define (decorate n x)
      (if (memq (node-class n) '(lambda global-ref local-ref))
          (typed n x)
          x))
    (pp (apply node-shape
               n #f
               (if (memq 'dump-with-types (or (verbose) '()))
                   (list decorate)
                   '())))
    n)

  ;; drop toplevel code

  (define (treeshake n)
    (case (node-class n)
      ((unspecified const global-ref local-ref lambda) #f)
      ((begin)
        (let ((ns (filter-map treeshake (node-subnodes n))))
          (cond ((null? ns) #f)
                ((null? (cdr ns)) (car ns))
                (else
                  (node-subnodes-set! n ns)
                  n))))
      ((call)
        (node-class-set! n 'begin)
        (node-data-set! n '())
        (treeshake n))
      ((let)
        (let* ((body (treeshake (last (node-subnodes n)))))
          (and body n)))
      (else n)))

  ;; final validation pass: we replace primitive call-id's with their
  ;; specialization if needed (as in the inference pass) or fail if
  ;; overloaded.
  (define (validate n)
    (let ((failed #f))
      (define (fail n fstr . args)
        (let ((loc (node-location n)))
          (elucidate
            (if loc
                (format "Error: ~a: ~?\n" loc fstr args)
                (format "Error: ~?\n" fstr args)))
          (set! failed #t)))
      (define (specialize n id name argt)
        (let ((sp (get id 'crunch:specialize)))
          (if sp
    		  (let ((sid (sp (map typename argt))))
    		    (or sid
    		        (begin
    		          (fail n "overloaded primitive `~a' can not be used as first class value" name)
    		  	      id)))
    		  id)))
      (define (walk n)
        (case (node-class n)
          ((const)
           (let ((x (first (node-data n))))
             (when (symbol? x)
               (let ((prim (get x 'crunch:primitive)))
                 (when prim
                   (if (get x 'crunch:overload)
                       (fail n "overloaded primitive `~a' can not be used as first class value" prim)
                       (let* ((tv (get prim 'crunch:type))
                              (loc (node-location n))
                              (argc (procedure-required-arguments tv))
                              (argt (procedure-argument-types tv))
                              (vs (map (lambda _ (gensym 'v)) argt))
                              (id (specialize n x prim argt)))
                         (replace-node!
                           n
                           (make-node
                             loc
                             'lambda (list (gensym 'f) argc vs vs #f)
                             (list
                              (make-node
                               loc
                               'idcall
                               (list id #t (< argc (length argt)))
                               (map (lambda (v tv)
                                      (make-node
                                        loc
                                        'local-ref (list v v)
                                        '()
                                        tv))
                                 vs argt)
                               (procedure-result-type tv)))
                             tv)))))))))
          (else (for-each walk (node-subnodes n)))))
      (walk n)
      (not failed)))

  (define pseudo-source-file-name
    (let ((count 0))
      (lambda ()
        (set! count (add1 count))
        (format "<forms ~a>" count))))

  ;; toplevel
  (define (crunch . args)
    (let ((all '())
          (cleanup #f)
          (eh (error-handler)))
      (parameterize ((exported '())
                     (whole-program #f)
                     (expanded-modules '())
                     (register-exports #t)
                     (error-handler
                       (lambda (ex)
                         (when cleanup (cleanup-node cleanup))
                         (eh ex))))
        (define (fail name)
          (crunch-error "compilation failed" name '()))
        (define (compile nodes fname)
          (let ((n0 (apply node* #f 'begin '() nodes)))
            (dribble "treeshake")
            (and-let* ((n1 (treeshake n0)))
              (set! cleanup n1)
              (when (memq 'expanded (or (verbose) '()))
                (dump n1 "expand"))
              (and-let* ((n2 (analyze n1)))
                (set! cleanup n2)
                (when (memq 'resolved (or (verbose) '()))
                  (dump n2 "resolve"))
                (and-let* ((n3 (optimize n2)))
                  (set! cleanup n3)
                  (dribble "validating")
                  (unless (validate n3)
                    ((node-dump-hook) n3 "ERROR: validation failed")
                    (fail fname))
                  (when (memq 'optimized (or (verbose) '()))
                    (dump n3 "optimize" #f))
                  ((node-dump-hook) n3 "* success\n")
                  (dribble "generating code")
                  (emit n3 (or (consumer) (file-consumer fname)))
                  (let ((ctb (compile-to-binary)))
                    (when ctb
                      (let ((st (system (format "~a ~a '~a' -o '~a'"
                                                c-compiler
                                                (apply conc
                                                       (map (cut conc " '" <> "'")
                                                         (cdr ctb)))
                                                (output-filename)
                                                (car ctb)))))
                        (unless (zero? st)
                          (elucidate
                            (format "compiling `~a' to '~a' failed with exit status ~a\n"
                                    (output-filename) (car ctb) st))
                          (fail fname)))))
                  #t)))))
        (define (compile-source fname in)
          (fluid-let ((##sys#read-error-with-line-number #t)
                      (##sys#current-load-filename fname)
                      (##sys#current-source-filename fname))
            (let ((cs (case-sensitive)))
              (dynamic-wind
                (lambda () (set! cs (case-sensitive)))
                (lambda ()
                  (let loop ((nodes '()))
                    (let ((x (read-with-source-info in)))
                      (cond ((eof-object? x)
                             (if (whole-program)
                                 (set! all (append nodes all))
                                 (unless (compile (reverse nodes) fname)
                                   (fail fname))))
                            (else
                              (let ((n (evaluate (crunch-expand x))))
                                (loop (cons n nodes))))))))
                (lambda ()
                  (case-sensitive cs)
                  (close-input-port in))))))
        (parse-args
          args
          (lambda (fnames)
            (when (null? fnames) (error 'crunch "no files or code given"))
            (when (debugger-active)
              (elucidate (format "arguments:\n~a\n"
                           (with-output-to-string (cut pp args)))))
            (for-each
              (lambda (src)
                (if (pair? src)
                    (if (debugger-active)
                        (let* ((fname (pseudo-source-file-name))
                               (txt (with-output-to-string (cut pp src))))
                          (compile-source fname (open-input-string txt))
                          (register-pseudo-source-file fname txt))
                        (let ((n (evaluate (crunch-expand src))))
                          (if (whole-program)
                              (set! all (append (list n) all))
                              (unless (compile (list n) #f)
                                (fail src)))))
                    (begin
                      (when (debugger-active)
                        (register-source-file src))
                      (compile-source src (open-input-file src)))))
              fnames)
            (cond ((whole-program)
                   (unless (compile (reverse all)
                                    (or (output-filename)
                                        (begin
                                          (error 'crunch "output filename must be specified")
                                          (fail #f))))
                     (fail #f)))
                  ((null? (expanded-modules))
                   ((or (export-consumer) void) #f
                    (map (lambda (exp)
                           (list (car exp) (cdr exp)
                                 (get (cdr exp) 'crunch:type)))
                      (exported)))
                   (when cleanup (cleanup-node cleanup)))
                  (else
                    (for-each
                      (lambda (m)
                        (let ((name (##sys#module-name m)))
                          ((or (export-consumer) void) name (module-exports m))
                          (when (and (not (debugger-active))
                          		     (or (all-import-libraries)
                                         (member name (import-libraries))))
                            (emit-import-lib name m))))
                      (expanded-modules))
                    (when cleanup (cleanup-node cleanup))))
            (set! ##sys#module-table
              (filter (lambda (ma) (not (memq (cdr ma) (expanded-modules))))
                      ##sys#module-table)))))))

  (define (file-consumer fname)
    (let ((out #f)
          (fname (or (output-filename)
                     (pathname-replace-extension fname "c"))))
      (lambda (code)
        (unless out
          (set! out
            (if (string=? "-" fname)
                (current-output-port)
                (open-output-file fname))))
        (if (eof-object? code)
            (unless (string=? "-" fname)
              (close-output-port out))
            (write-string code out)))))

  ;; import library handling:

  (define (import-library-name name)
    (string-append (symbol->string name) ".import.scm"))

  (define (emit-import-lib name mod)
    (let* ((imps
             `((begin
                ,@(##sys#compiled-module-registration mod #f)
                ,(declarations mod))))
           (fname (import-library-name name))
           (oldimps
             (and (file-exists? fname)
                  (call-with-input-file fname read-expressions))))
      (cond ((equal? imps oldimps)
             (when (verbose)
               (elucidate (format "not generating import library `~a' for module `~a' because imports did not change\n"
               			    fname name))))
            (else
              (when (verbose)
                (elucidate (format "generating import library `~a' for module `~a' ...\n"
                			 fname name)))
              (with-output-to-file fname
                (lambda ()
                  (print ";;;; " fname " - GENERATED BY CRUNCH "
                         (crunch-version) " -*- Scheme -*-\n")
                  (for-each pretty-print imps)
                  (print "\n;; END OF FILE"))))) ) )

  (define (read-expressions #!optional (port (current-input-port)))
    (do ((x (read port) (read port))
         (i 0 (add1 i))
         (xs '() (cons x xs)))
        ((eof-object? x) (reverse xs))))

  (define (module-exports mod)
    (receive (_ ve _) (##sys#module-exports mod)
        (map (lambda (e)
               (let* ((en (cdr e))
                      (t (get en 'crunch:type)))
                 (list (car e) en t)))
          ve)))

  (define (declarations mod)
    (let ((regs (map (match-lambda
                       ((e en t)
                        (let ((tn (typename t)))
                          `(crunch.compiler#declare-entity
                             ',en ',tn
                             ',(match (typevar-value-set t)
                                 (((? symbol? id))
                                  (get id 'crunch:external-name))
                                 (_ #f))))))
                  (module-exports mod))))
      `(if (memq #:crunched ##sys#features) (begin ,@regs))))

  (define (module-rename sym prefix)
    (string->symbol
      (string-append
        (symbol->string prefix)
        "#"
        (symbol->string sym) ) ) )

  (define (declare-entity name t #!optional extname)
    (put! name 'crunch:type (typespec t extname)))

  (set! ##sys#features (cons #:crunched ##sys#features))

)

